/*
Bounce 0.9
Copyright 2003 Tim Gradwell
Bounce is distributed under the terms of the GNU General Public License 

This file is part of Bounce.

Bounce is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

Bounce is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Bounce; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "game.h"

void launchBall(float colX)
{
   int x_displacement;

   x_displacement = (int)ballX+(int)ballWidth - (int)colX;

   if (x_displacement < 6)
      ballXVelocity = -ballVelocity / 0.700f;

   else if (x_displacement < 13)
      ballXVelocity = -ballVelocity / 1.000f;

   else if (x_displacement < 19)
      ballXVelocity = -ballVelocity / 1.428f;

   else if (x_displacement < 26)
      ballXVelocity = -ballVelocity / 2.145f;

   else if (x_displacement < 33)
      ballXVelocity = -ballVelocity / 3.732f;
   
   else if (x_displacement < 40)
      ballXVelocity = -ballVelocity / 11.430f;

   else if (x_displacement < 47)
      ballXVelocity =  ballVelocity / 11.430f;

   else if (x_displacement < 54)
      ballXVelocity =  ballVelocity / 3.732f;

   else if (x_displacement < 61)
      ballXVelocity =  ballVelocity / 2.145f;

   else if (x_displacement < 67)
      ballXVelocity =  ballVelocity / 1.428f;

   else if (x_displacement < 74)
      ballXVelocity =  ballVelocity / 1.000f;

   else
      ballXVelocity =  ballVelocity / 0.700f;

   ballDownwards=false;

   if (ballXVelocity > 0)
      ballLeftwards = false;
   else
      ballLeftwards = true;

   holdingBall=false;
}

void endOfLife()
{
   lives--;
   if (lives < 1)
      gameOver();
   playSound(SOUND_LOOSE_LIFE);
   holdingBall = true;
}

void bounceWalls()
{
   if (newBallX < LEFT_BOUND)
   {
      ballXVelocity = -ballXVelocity;
      newBallX = 2*LEFT_BOUND - newBallX;
      ballLeftwards = false;
      playSound(SOUND_WALL);
   }

   if (newBallX+ballWidth > RIGHT_BOUND)
   {
      ballXVelocity = -ballXVelocity;
      newBallX = 2*RIGHT_BOUND - newBallX - 2*ballWidth;
      ballLeftwards = true;
      playSound(SOUND_WALL);
   }

   if (newBallY+ballHeight > TOP_BOUND)
   {
      ballYVelocity = - ballYVelocity;
      newBallY = 2*TOP_BOUND - newBallY - 2*ballHeight;
      ballDownwards = true;
      playSound(SOUND_WALL);
   }

   if (newBallY < BOTTOM_BOUND)
/*   {
      ballYVelocity = - ballYVelocity;
      ballDownwards = false;
   }
*/
      endOfLife();
}

void bounceBat()
{
   float xIntersect;
   if ((newBallY+ballHeight >= batY) &&(newBallY < (float)(batY+batHeight)))
   {
      xIntersect = (int)((((float)batY-ballY)*newBallX + (newBallY-(float)batY)*ballX) / (newBallY-ballY));
      if ((xIntersect+ballWidth >= batX)&&(xIntersect < batX+batWidth))
      {
         newBallY=2*(batY+batHeight)-newBallY;
         ballYVelocity=-ballYVelocity;
         playSound(SOUND_BAT);
         launchBall(batX);
      }
      if (powerupEffect[3].effect &&
         ((xIntersect+ballWidth >= uberBatX)&&(xIntersect < uberBatX+batWidth)))
      {
         newBallY=2*(batY+batHeight)-newBallY;
         ballYVelocity=-ballYVelocity;
         playSound(SOUND_BAT);
         launchBall(uberBatX);
      }

   }
}

void brickHit(int brickX, int brickY)
{
   score += 50;

   if (((int)batX-brickX-brickY) % 5 == 0)
      generatePowerup(brickX, brickY);

   if (brick[brickX][brickY] > 7-difficulty)
   {
      playSound(SOUND_BRICK_CHANGE_COLOR);
      brick[brickX][brickY] -= 4;
   }
   else
   {
      playSound(SOUND_BRICK);
      brick[brickX][brickY] = 0;
   }
}

void bounceBrickLeft(int brickX, int brickY)
{
   float brickYCoord = brickCoord[brickX][brickY][0][1];
   float brickXCoord = brickCoord[brickX][brickY][0][0];
   if (brick[brickX][brickY])
   {
      if ((newBallX + ballWidth) > brickXCoord && (ballX + ballWidth) <= brickXCoord)
      {
         float yIntersect, x1, x2, Y, R;

         x1 = brickXCoord-(ballX+ballWidth);
         x2 = (newBallX+ballWidth)-brickXCoord;

         Y = newBallY - ballY;

         R = x1 / x2;

         yIntersect = ((R*Y)/(1+R))+ballY;
         
         if (yIntersect /*measured from bottom right corner of ball*/ >= brickYCoord-ballHeight && yIntersect < brickYCoord+brickHeight)
         {
            if (!powerupEffect[1].effect)
            {
               newBallX = 2*brickXCoord - ballX - 2*ballWidth;
               ballLeftwards = true;
            }
            brickHit(brickX, brickY);
         }
      }
   }
}

void bounceBrickRight(int brickX, int brickY)
{
   float brickYCoord = brickCoord[brickX][brickY][0][1]; /* bottom-left's Y */
   float brickXCoord = brickCoord[brickX][brickY][1][0]; /* top-right's X */
   if (brick[brickX][brickY])
   {
      if (newBallX < brickXCoord && ballX >= brickXCoord)
      {
         float yIntersect, x1, x2, Y, R;

         x1 = ballX-brickXCoord;
         x2 = brickXCoord - newBallX;

         Y = newBallX - ballX;

         R = x1 / x2;

         yIntersect = ((R*Y)/(1+R))+ballY;
         
         if (yIntersect /*measured from bottom left corner of ball*/ > brickYCoord-ballHeight && yIntersect <= brickYCoord+brickHeight)
         {
            if (!powerupEffect[1].effect)
            {
               newBallX = 2*brickXCoord - ballX;
               ballLeftwards = false;
            }
            brickHit(brickX, brickY);
         }
      }
   }
}

void bounceBrickTop(int brickX, int brickY)
{
   float brickYCoord = brickCoord[brickX][brickY][1][1]; /* top-right's y coord */
   float brickXCoord = brickCoord[brickX][brickY][0][0]; /* bottom-left's x coord */
   if (brick[brickX][brickY])
   {
      if (newBallY < brickYCoord && ballY >= brickYCoord)
      {
         float xIntersect, y1, y2, X, R;

         y1 = ballY-brickYCoord;
         y2 = brickYCoord - newBallY;

         X = newBallX - ballX;

         R = y1 / y2;

         xIntersect = ((R*X)/(1+R))+ballX;
         
         if (xIntersect /*measured from bottom left corner of ball*/ >= brickXCoord-ballWidth && xIntersect < brickXCoord+brickWidth)
         {
            if (!powerupEffect[1].effect)
            {
               newBallY = 2*brickYCoord - ballY;
               ballDownwards = false;
            }
            brickHit(brickX, brickY);
         }
      }
   }
}

void bounceBrickBottom(int brickX, int brickY)
{
   float brickYCoord = brickCoord[brickX][brickY][0][1]; /* bottom-left's y coord */
   float brickXCoord = brickCoord[brickX][brickY][0][0]; /* bottom-left's x coord */
   if (brick[brickX][brickY])
   {
      if ((newBallY + ballHeight) > brickYCoord && (ballY + ballHeight) <= brickYCoord)
      {
         float xIntersect, y1, y2, X, R;

         y1 = brickYCoord-(ballY+ballHeight);
         y2 = (newBallY+ballHeight)-brickYCoord;

         X = newBallX - ballX;

         R = y1 / y2;

         xIntersect = ((R*X)/(1+R))+ballX;
         
         if (xIntersect /*measured from top left corner of ball*/ >= brickXCoord-ballWidth && xIntersect < brickXCoord+brickWidth)
         {
            if (!powerupEffect[1].effect)
            {
               newBallY = 2*brickYCoord - ballY - 2*ballHeight;
               ballDownwards = true;
            }
            brickHit(brickX, brickY);
         }
      }
   }
}

void bounceBrick()
{

   int initX, initY, finalX, finalY, incX, incY;
   int colX, colY;

   if (ballXVelocity > 0)
   {
      initX = 0;
      finalX = COLUMNS;
      incX = 1;
   }
   else
   {
      initX = COLUMNS-1;
      finalX = -1;
      incX = -1;
   }

   if (ballYVelocity > 0)
   {
      initY = ROWS-1;
      finalY = -1;
      incY = -1;
   }
   else
   {
      initY = 0;
      finalY = ROWS;
      incY = 1;
   }

   for (colX = initX; colX != finalX; colX+=incX)
      for (colY = initY; colY != finalY; colY+=incY)
      {
         if (brick[colX][colY])
         {
            if (incX == 1)
            {
               bounceBrickLeft(colX, colY);
               bounceBrickRight(colX, colY);
            }
            else
            {
               bounceBrickRight(colX, colY);
               bounceBrickLeft(colX, colY);
            }

            if (incY == 1)
            {
               bounceBrickTop(colX, colY);
               bounceBrickBottom(colX, colY);
            }
            else
            {
               bounceBrickBottom(colX, colY);
               bounceBrickTop(colX, colY);
            }
         }
      }
   checkEndOfLevel();
}

void checkEndOfLevel()
{
   int i, j, bricks=0;
   for (i=0; i < ROWS; i++)
      for (j=0; j < COLUMNS; j++)
         bricks += brick[j][i];
   if (bricks == 0)
   {
      playSound(SOUND_LEVEL_UP);
      newLevel();
   }
}

void predictBallPosition()
{
   if (ballDownwards&& ballYVelocity > 0)
      ballYVelocity = -ballYVelocity;
   if (!ballDownwards && ballYVelocity < 0)
      ballYVelocity = -ballYVelocity;

   if (ballLeftwards&& ballXVelocity > 0)
      ballXVelocity = -ballXVelocity;
   if (!ballLeftwards && ballXVelocity < 0)
      ballXVelocity = -ballXVelocity;

   setBallVelocity();

   newBallX=ballX+ballXVelocity*ballVelocity*slowdownFactor*GAME_SPEED*timeElapsed;
   newBallY=ballY+ballYVelocity*ballVelocity*slowdownFactor*GAME_SPEED*timeElapsed;
}

void updateBallPosition()
{   
   ballX=newBallX;
   ballY=newBallY;
}

void scorePoints(int value)
{
   score +=10*value;
}

void newGame()
{
   batX = 500;
   score = 0;
   lives = 6 - difficulty;
   level = 0;
   gameSection = SECTION_GAME;
   newLevel();
}

void newLevel()
{
   int i, j;

   level++;
   ballYVelocity=2.0;
   setBallVelocity();
   ballDownwards = false;
   ballLeftwards = false;
   holdingBall = true;

   for (i = 0; i < ROWS; i++)
   {
      for (j = 0; j < COLUMNS; j++)
         brick[j][i] = levelBrick[(level-1)%10][j][i];
   }
   
   if ((level-1) / 10 > 0)
      for (i = 0; i < ROWS; i++)
      {
         for (j = 0; j < COLUMNS; j++)
            if (brick[j][i] < 7 && brick[j][i] > 0)
               brick[j][i]+= level / 10;
      }

   powerups = 0;
   for (i=0; i < MAX_POWERUPS; i++)
      powerup[i].color = 0;

   powerupEffects = 0;
   for (i=1; i < 8; i++)
   {
      powerupEffect[i].effect = false;
      powerupEffect[i].time = 0;
      powerupEffect[i].duration = 500;
   }

   slowdown = 100.0;
   slowdownFactor = 1.0;

}

void generatePowerup(int brickX, int brickY)
{
   int i;

   if (powerups < MAX_POWERUPS)
   {
      i = 0;
      while(powerup[i].color > 0)
         i++;

      powerup[i].color = ((int)ballX % 7) + 1;
      powerup[i].x = brickCoord[brickX][brickY][0][0] + brickWidth / 2.0;
      powerup[i].y = brickCoord[brickX][brickY][1][1];
      powerups++;
   }
}

void powerupCatch()
{
   int i;
   for (i = 0; i < MAX_POWERUPS; i++)
      if (powerup[i].color > 0)
         if (powerup[i].y < batY+batHeight+POWERUP_SIZE/2.0)
            if (powerup[i].x+POWERUP_SIZE/2.0 > batX && powerup[i].x-POWERUP_SIZE/2.0 < batX+batWidth)
               powerupactivate(i);
}

void powerupactivate(int p)
{
   int p2 = powerup[p].color;

   score += 100;

   if (!powerupEffect[powerup[p].color].effect)
   {
      powerupEffects++;
      powerupEffect[p2].effect = true;
   }
   switch(p2)
   {
   case 1:
      playSound(SOUND_POWERBALL);
      break;
   case 2: /* ORANGE MORE SLOW DOWN */
      playSound(SOUND_SLOW_DOWN);
      slowdown = 100.0;
      break;
   case 3:
      playSound(SOUND_UBER_BAT);
      break;
   case 4: /* GREEN EXTRA LIFE */
      playSound(SOUND_EXTRA_LIFE);
      lives++;
      break;
   case 5:
      playSound(SOUND_SPEED_UP);
      break;
   case 6:
      playSound(SOUND_BIG_BAT);
      break;
   case 7:
      playSound(SOUND_SMALL_BAT);
      break;
   }
   powerup[p].color = 0;
   powerups--;
}

void setBallVelocity()
{
   switch(difficulty)
   {
   case(DIFFICULTY_EASY):
      ballVelocity = BALL_VELOCITY_EASY;
      break;
   case(DIFFICULTY_NORMAL):
      ballVelocity = BALL_VELOCITY_NORMAL;
      break;
   case(DIFFICULTY_HARD):
      ballVelocity = BALL_VELOCITY_HARD;
      break;
   }   

   if (powerupEffect[5].effect)
      ballVelocity*=SPEED_UP;
}

void gameOver()
{
   int i;
   
   replace = 5;

   playSound(SOUND_GAME_OVER);

   gameSection = SECTION_GAME_OVER;

   if (score > highscores[4])
   {
      replace = 4;
      for (i = 3; i >= 0; i--)
         if (score > highscores[i])
         {
            highscores[i+1] = highscores[i];
            highlevels[i+1] = highlevels[i];
            strcpy(names[i+1], names[i]);
            replace = i;
         }
   }
   
   if (replace == 5)
      gameOverSection = GAME_OVER_LOOSE;
   else
   {
      cursor = 0;
      highscores[replace] = score;
      highlevels[replace] = level;
      strcpy(names[replace],"   ");
      gameOverSection = GAME_OVER_HIGH_SCORE;
   }
}

